from contextlib import suppress
from typing import overload, Literal, TypeAlias, AsyncIterator, Callable
from datetime import datetime
from re import search as re_search

from aiohttp import ClientResponseError
from aiohttp.client import _RequestContextManager

from ..typed import WalletInfo
from ..constants import (
    STEAM_URL,
    App,
    AppContext,
    MarketListingStatus,
    MarketHistoryEventType,
    T_PARAMS,
    T_PAYLOAD,
    T_HEADERS,
    EResult,
    Currency,
)
from ..models import (
    EconItem,
    MyMarketListing,
    BuyOrder,
    ItemDescription,
    MarketListingItem,
    MarketHistoryEvent,
    MarketHistoryListing,
    MarketHistoryListingItem,
    PriceHistoryEntry,
    MarketListing,
)
from ..helpers import currency_required
from ..exceptions import EResultError, SessionExpired
from ..utils import create_ident_code, buyer_pays_to_receive, calc_market_listing_fee
from .public import SteamCommunityPublicMixin, T_SHARED_DESCRIPTIONS
from .confirmation import ConfirmationMixin


MY_LISTINGS_DATA: TypeAlias = tuple[list[MyMarketListing], list[MyMarketListing], list[BuyOrder], int]
MY_MARKET_HISTORY_DATA: TypeAlias = tuple[list[MarketHistoryEvent], int]


class MarketMixin(ConfirmationMixin, SteamCommunityPublicMixin):
    """
    Mixin with market related methods.
    Depends on `ConfirmationMixin`, `SteamCommunityPublicMixin`.
    """

    __slots__ = ()

    # nice list
    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        price: int,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        price: int,
        confirm: Literal[False] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> None:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        to_receive: int,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        to_receive: int,
        confirm: Literal[False] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> None:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        price: int,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        price: int,
        confirm: Literal[False] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> None:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        to_receive: int,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        to_receive: int,
        confirm: Literal[False] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> None:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        price: int,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> MyMarketListing:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: EconItem,
        *,
        to_receive: int,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> MyMarketListing:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        price: int,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> MyMarketListing:
        ...

    @overload
    async def place_sell_listing(
        self,
        obj: int,
        app_context: AppContext,
        *,
        to_receive: int,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> MyMarketListing:
        ...

    async def place_sell_listing(
        self,
        obj: EconItem | int,
        app_context: AppContext = None,
        *,
        price: int = None,
        to_receive: int = None,
        fetch=False,
        confirm=True,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> int | MyMarketListing | None:
        """
        Create and place sell listing.
        If `confirm` is `True` - return listing id of created and confirmed sell listing,
        if this requires confirmation. If not - return `None`.

        .. note::
            * Money should be only and only in account wallet currency
            * `price` or `to_receive` is integers equal to cents

        :param obj: `EconItem` that you want to list on market or asset id
        :param app_context: `Steam` app+context
        :param price: money that buyer must pay. Include fees
        :param to_receive: money that you want to receive
        :param fetch: make request and return a listing
        :param confirm: confirm listing or not if steam demands mobile confirmation
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :return: sell listing id, `MyMarketListing` or `None`
        :raises TypeError:
        :raises EResultError: for ordinary reasons
        """

        if isinstance(obj, EconItem):
            asset_id = obj.asset_id
            app_context = obj.app_context
        else:
            asset_id = obj

        # prevent user from mistake and potentially money loss
        if to_receive and price:
            raise TypeError("The `price` and `to_receive` arguments are mutually exclusive!")
        elif type(price) is float or type(to_receive) is float:
            raise TypeError(
                "The `price` and `to_receive` arguments should be integers. Did you forget to convert price to cents?"
            )

        to_receive = to_receive or buyer_pays_to_receive(price)[2]

        data = {
            "assetid": asset_id,
            "sessionid": self.session_id,
            "contextid": app_context.context,
            "appid": app_context.app.value,
            "amount": 1,
            "price": to_receive,
            **payload,
        }
        headers = {"Referer": str(STEAM_URL.COMMUNITY / f"profiles/{self.steam_id}/inventory"), **headers}
        r = await self.session.post(STEAM_URL.MARKET / "sellitem/", data=data, headers=headers)
        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to place sell listing"), success, rj)

        to_return = None

        if rj.get("needs_mobile_confirmation") and confirm:
            conf = await self.confirm_sell_listing(asset_id, app_context)
            to_return = conf.creator_id

        if fetch:
            to_return = await self.get_my_sell_listing(asset_id=asset_id, _item_descriptions_map=_item_descriptions_map)

        return to_return

    @overload
    async def get_my_sell_listing(
        self,
        obj: int = ...,
        *,
        price: int = ...,
        need_confirmation: bool = ...,
        asset_id: int = ...,
        market_hash_name: str = ...,
        ident_code: str = ...,
        app_context: AppContext = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
        **listing_item_attrs,
    ) -> MyMarketListing | None:
        ...

    @overload
    async def get_my_sell_listing(
        self,
        obj: Callable[[MyMarketListing], bool],
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> MyMarketListing | None:
        ...

    async def get_my_sell_listing(
        self,
        obj: int | Callable[[MyMarketListing], bool] = None,
        *,
        price: int = None,
        need_confirmation: bool = False,
        asset_id: int = None,
        market_hash_name: str = None,
        ident_code: str = None,
        app_context: AppContext = None,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        **listing_item_attrs,
    ) -> MyMarketListing | None:
        """
        Fetch and iterate over sell listings pages until find one that satisfies passed arguments.

        :param obj: listing id or predicate function
        :param price: listing price
        :param need_confirmation: get listing from `to_confirm` list
        :param asset_id: asset id of listing item
        :param market_hash_name: market hash name of listing item
        :param ident_code: ident code of listing item
        :param app_context: `Steam` app+context
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :param listing_item_attrs: additional listing item attributes and values
        :return: `MyMarketListing` or `None`
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        if callable(obj):
            predicate = obj
        else:

            def predicate(l: MyMarketListing):
                if obj is not None and l.id != obj:
                    return False
                if price is not None and l.price != price:
                    return False
                if asset_id is not None and l.item.asset_id != asset_id:
                    return False
                if market_hash_name is not None and l.item.description.market_hash_name != market_hash_name:
                    return False
                if ident_code is not None and l.item.id != ident_code:
                    return False
                if app_context is not None and l.item.app_context is not app_context:
                    return False

                for attr, value in listing_item_attrs.items():
                    if getattr(l.item, attr, None) != value:
                        return False

                return True

        async for listings, to_confirm, _, _ in self.my_listings(
            params=params,
            headers=headers,
            _item_descriptions_map=_item_descriptions_map,
        ):
            with suppress(StopIteration):
                return next(filter(predicate, to_confirm if need_confirmation else listings))

    def cancel_sell_listing(
        self,
        obj: MyMarketListing | int,
        *,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
    ) -> _RequestContextManager:
        """
        Simply cancel sell listing.

        :param obj: `MyMarketListing` or listing id
        :param payload: extra payload data
        :param headers: extra headers to send with request
        """

        listing_id: int = obj.id if isinstance(obj, MyMarketListing) else obj
        data = {"sessionid": self.session_id, **payload}
        headers = {"Referer": str(STEAM_URL.MARKET), **headers}
        return self.session.post(STEAM_URL.MARKET / f"removelisting/{listing_id}", data=data, headers=headers)

    @overload
    async def place_buy_order(
        self,
        obj: ItemDescription,
        *,
        price: int,
        quantity: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_buy_order(
        self,
        obj: str,
        app: App,
        *,
        price: int,
        quantity: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def place_buy_order(
        self,
        obj: ItemDescription,
        *,
        price: int,
        quantity: int = ...,
        fetch: Literal[True] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> BuyOrder:
        ...

    @overload
    async def place_buy_order(
        self,
        obj: str,
        app: App,
        *,
        price: int,
        quantity: int = ...,
        fetch: Literal[True] = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> BuyOrder:
        ...

    @currency_required
    async def place_buy_order(
        self,
        obj: str | ItemDescription,
        app: App = None,
        *,
        price: int,
        quantity=1,
        fetch=False,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> int | BuyOrder:
        """
        Place buy order on market.

        :param obj: `ItemDescription` or market hash name
        :param app: `Steam` app if `obj` is market hash name
        :param price: price of single item
        :param quantity: just quantity
        :param fetch: make request and return buy order
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :return: buy order id or `BuyOrder`
        :raises EResultError: for ordinary reasons
        """

        if isinstance(obj, ItemDescription):
            name = obj.market_hash_name
            app = obj.app
        else:
            name = obj

        data = {
            "sessionid": self.session_id,
            "currency": self.currency.value,
            "appid": app.value,
            "market_hash_name": name,
            "price_total": price * quantity,
            "quantity": quantity,
            **payload,
        }
        headers = {"Referer": str(STEAM_URL.MARKET / f"listings/{app.value}/{name}"), **headers}
        r = await self.session.post(STEAM_URL.MARKET / "createbuyorder/", data=data, headers=headers)
        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to create buy order"), success, rj)

        to_return = int(rj["buy_orderid"])
        if fetch:
            to_return = await self.get_my_buy_order(to_return, _item_descriptions_map=_item_descriptions_map)

        return to_return

    @overload
    async def get_my_buy_order(
        self,
        obj: int = ...,
        *,
        price: int = ...,
        quantity: int = ...,
        market_hash_name: str = ...,
        ident_code: str = ...,
        app: App = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
        **item_description_attrs,
    ) -> BuyOrder | None:
        ...

    @overload
    async def get_my_buy_order(
        self,
        obj: Callable[[BuyOrder], bool],
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> BuyOrder | None:
        ...

    async def get_my_buy_order(
        self,
        obj: int | Callable[[BuyOrder], bool] = None,
        *,
        price: int = None,
        quantity: int = None,
        market_hash_name: str = None,
        ident_code: str = None,
        app: App = None,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        **item_description_attrs,
    ) -> BuyOrder | None:
        """
        Fetch and iterate over buy order pages until find one that satisfies passed arguments.

        :param obj: order id or predicate function
        :param price: order price
        :param quantity: order quantity
        :param market_hash_name: name of item description
        :param ident_code: ident code of item description
        :param app: `Steam` app
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :param item_description_attrs: additional item description attributes and values
        :return: `BuyOrder` or `None`
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        if callable(obj):
            predicate = obj
        else:

            def predicate(o: BuyOrder):
                if obj is not None and o.id != obj:
                    return False
                if price is not None and o.price != price:
                    return False
                if quantity is not None and o.quantity != quantity:
                    return False
                if market_hash_name is not None and o.item_description.market_hash_name != market_hash_name:
                    return False
                if ident_code is not None and o.item_description.id != ident_code:
                    return False
                if app is not None and o.item_description.app is not app:
                    return False

                for attr, value in item_description_attrs.items():
                    if getattr(o.item_description, attr, None) != value:
                        return False

                return True

        # count 1 to reduce unnecessary work, better check this
        data = await self.get_my_listings(
            count=1,
            params=params,
            headers=headers,
            _item_descriptions_map=_item_descriptions_map,
        )
        return next(filter(predicate, data[2]), None)

    async def cancel_buy_order(self, order: int | BuyOrder, *, payload: T_PAYLOAD = {}, headers: T_HEADERS = {}):
        """
        Just cancel buy order.

        :param order: `BuyOrder` or buy order id
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :raises EResultError: for ordinary reasons
        """

        if isinstance(order, BuyOrder):
            order_id = order.id
        else:
            order_id = order

        data = {"sessionid": self.session_id, "buy_orderid": order_id, **payload}
        headers = {"Referer": str(STEAM_URL.MARKET), **headers}
        r = await self.session.post(STEAM_URL.MARKET / "cancelbuyorder/", data=data, headers=headers)
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to cancel buy order"), success, rj)

    async def get_my_listings(
        self,
        *,
        start: int = 0,
        count: int = 100,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> MY_LISTINGS_DATA:
        """
        Fetch users market listings.

        .. note:: you can paginate active listings by yourself passing `start` arg

        :param start: start index
        :param count: listings per page. `Steam` do not accept value greater than 100
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: active listings, listings to confirm, buy orders, total count of active listings
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        params = {"norender": 1, "start": start, "count": count, **params}

        try:
            r = await self.session.get(STEAM_URL.MARKET / "mylistings", params=params, headers=headers)
        except ClientResponseError as e:
            raise SessionExpired if e.status == 400 else e  # Are we sure that there status code 400 and not 403?

        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch user listings"), success, rj)

        # no need to check `assets` or `total_count`

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        self._parse_item_descrs_from_my_listings(rj, _item_descriptions_map)

        active = self._parse_my_listings(rj["listings"], _item_descriptions_map)
        to_confirm = self._parse_my_listings(rj["listings_to_confirm"], _item_descriptions_map)
        buy_orders = self._parse_buy_orders(rj["buy_orders"], _item_descriptions_map)

        return active, to_confirm, buy_orders, rj["num_active_listings"]

    async def my_listings(
        self,
        *,
        start: int = 0,
        count: int = 100,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> AsyncIterator[list[MyMarketListing]]:
        """
        Fetch users market listings. Return async iterator to paginate over listing pages.

        .. note:: paginates only over active listing pages.
            If you need to get orders or listings to confirm use `get_my_listings` method

        :param start: start index
        :param count: listings per page. Steam do not accept value greater than 100
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of active `MyMarketListing`
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        more_listings = True
        while more_listings:
            # avoid excess destructuring
            listings_data = await self.get_my_listings(
                start=start,
                count=count,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
            )
            start += count
            more_listings = listings_data[3] > start

            yield listings_data[0]

    @classmethod
    def _parse_item_descrs_from_my_listings(
        cls,
        data: dict[str, dict | list[dict]],
        item_descrs_map: T_SHARED_DESCRIPTIONS,
    ):
        # assets field, has descriptions for listings, so we can not parse descrs from listings
        for app_id, app_data in (data["assets"] or {}).items():  # thanks Steam for an empty list instead of a dict
            for context_id, context_data in app_data.items():
                for asset_id, mixed_data in context_data.items():
                    key = create_ident_code(mixed_data["instanceid"], mixed_data["classid"], app_id)
                    if key not in item_descrs_map:
                        item_descrs_map[key] = cls._create_item_descr(mixed_data)

        for listing_data in data["listings_to_confirm"]:
            mixed_data = listing_data["asset"]
            key = create_ident_code(mixed_data["instanceid"], mixed_data["classid"], mixed_data["appid"])
            if key not in item_descrs_map:
                item_descrs_map[key] = cls._create_item_descr(mixed_data)

        for order_data in data["buy_orders"]:
            descr_data = order_data["description"]
            key = create_ident_code(descr_data["instanceid"], descr_data["classid"], descr_data["appid"])
            if key not in item_descrs_map:
                item_descrs_map[key] = cls._create_item_descr(descr_data)

    def _parse_my_listings(self, listings: list[dict], item_descrs_map: T_SHARED_DESCRIPTIONS) -> list[MyMarketListing]:
        return [
            MyMarketListing(
                id=int(l_data["listingid"]),
                price=l_data["price"],
                lister_steam_id=self.steam_id,
                time_created=datetime.fromtimestamp(l_data["time_created"]),
                item=MarketListingItem(
                    asset_id=int(l_data["asset"]["id"]),
                    unowned_id=int(l_data["asset"]["unowned_id"]) if "unowned_id" in l_data["asset"] else None,
                    owner_id=self.steam_id,
                    market_id=int(l_data["listingid"]),
                    unowned_context_id=int(l_data["asset"]["unowned_contextid"])
                    if "unowned_contextid" in l_data["asset"]
                    else None,
                    amount=int(l_data["asset"]["amount"]),
                    app_context=AppContext((App(int(l_data["asset"]["appid"])), int(l_data["asset"]["contextid"]))),
                    description=item_descrs_map[
                        create_ident_code(
                            l_data["asset"]["instanceid"],
                            l_data["asset"]["classid"],
                            l_data["asset"]["appid"],
                        )
                    ],
                ),
                status=MarketListingStatus(l_data["status"]),
                active=bool(l_data["active"]),
                item_expired=l_data["item_expired"],
                cancel_reason=l_data["cancel_reason"],
                time_finish_hold=l_data["time_finish_hold"],
            )
            for l_data in listings
        ]

    @classmethod
    def _parse_buy_orders(cls, orders: list[dict], item_descrs_map: T_SHARED_DESCRIPTIONS) -> list[BuyOrder]:
        return [
            BuyOrder(
                id=int(o_data["buy_orderid"]),
                price=int(o_data["price"]),
                item_description=item_descrs_map[
                    create_ident_code(
                        o_data["description"]["instanceid"],
                        o_data["description"]["classid"],
                        o_data["description"]["appid"],
                    )
                ],
                quantity=int(o_data["quantity"]),
                quantity_remaining=int(o_data["quantity_remaining"]),
            )
            for o_data in orders
        ]

    @overload
    async def buy_market_listing(
        self,
        obj: MarketListing,
        *,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> WalletInfo:
        ...

    @overload
    async def buy_market_listing(
        self,
        obj: int,
        price: int,
        market_hash_name: str,
        app: App,
        *,
        fee: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> WalletInfo:
        ...

    @currency_required
    async def buy_market_listing(
        self,
        obj: int | MarketListing,
        price: int = None,
        market_hash_name: str = None,
        app: App = None,
        *,
        fee: int = None,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
    ) -> WalletInfo:
        """
        Buy item listing from market.
        Unfortunately, `Steam` requires `referer` header to buy item,
        so `market_hash_name` and `app` is mandatory args in case of `obj` being a market id.

        .. note:: Make sure that listing converted currency same as wallet currency!

        :param obj: id for listing itself (aka market id) or `MarketListing`
        :param price: Can be found on listing data in `Steam` under field `converted_price`
        :param market_hash_name: market hash name of item. Needed for `Referer` header value
        :param app: `Steam` app
        :param fee: if fee of listing is different from default one,
            can be found on listing data in `Steam` under field `converted_fee`.
            If you don't know what this is - then you definitely do not need it
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :return: wallet info
        :raises EResultError: for regular reasons
        :raises ValueError: if converted currency of `MarketListing` is different from wallet currency
        """

        if isinstance(obj, MarketListing):
            if obj.converted_currency is None:
                raise ValueError("Listing must contain converted currency!")
            elif obj.converted_currency is not self.currency:
                raise ValueError(
                    f"Currency of listing ({obj.converted_currency.value}) is "
                    f"different from wallet ({self.currency.value}) one!"
                )

            listing_id = obj.id
            price = obj.converted_price
            fee = obj.converted_fee
            market_hash_name = obj.item.description.market_hash_name
            app = obj.item.app_context.app
        else:
            listing_id = obj

            if app is None or market_hash_name is None or price is None:
                raise ValueError("`app`, `market_hash_name` and `price` arguments must be provided")

        fee = fee or calc_market_listing_fee(price)
        data = {
            "sessionid": self.session_id,
            "currency": self.currency.value,
            "subtotal": price,
            "fee": fee,
            "total": price + fee,
            "quantity": 1,
            **payload,
        }
        headers = {"Referer": str(STEAM_URL.MARKET / f"listings/{app.value}/{market_hash_name}"), **headers}
        r = await self.session.post(
            STEAM_URL.MARKET / f"buylisting/{listing_id}",
            data=data,
            headers=headers,
            raise_for_status=False,
        )
        # ClientResponseError with code 502 [Bad Gateway] will be raised in case of insufficient balance, need
        # to do something with this somehow ...
        rj: dict[str, dict[str, str]] = await r.json()
        wallet_info: WalletInfo = rj.get("wallet_info", {})
        success = EResult(wallet_info.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", wallet_info.get("message", "Failed to buy listing")), success, rj)

        # how about to return remaining balance only?
        return wallet_info

    async def get_my_market_history(
        self,
        *,
        start: int = 0,
        count: int = 100,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        _market_history_econ_items_map: dict[str, MarketHistoryListingItem] = None,
        _market_history_listings_map: dict[int, MarketHistoryListing] = None,
    ) -> MY_MARKET_HISTORY_DATA:
        """
        Fetch market history of self.

        .. note:: You can paginate by yourself passing `start` arg

        :param start: start index
        :param count: listings per page. Steam do not accept value greater than 100
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `MarketHistoryEvent`, total_count
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        params = {"norender": 1, "start": start, "count": count, **params}

        try:
            r = await self.session.get(STEAM_URL.MARKET / "myhistory", params=params, headers=headers)
        except ClientResponseError as e:
            raise SessionExpired if e.status == 400 else e

        rj: dict = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch user listings"), success, rj)

        if not rj["total_count"] or not rj["assets"]:  # safe
            return [], 0

        if _item_descriptions_map is None:
            _item_descriptions_map = {}
        if _market_history_econ_items_map is None:
            _market_history_econ_items_map = {}
        if _market_history_listings_map is None:
            _market_history_listings_map = {}

        self._parse_item_descrs_from_my_listings(rj["assets"], _item_descriptions_map)
        self._parse_assets_for_history_listings(rj["assets"], _item_descriptions_map, _market_history_econ_items_map)
        self._parse_history_listings(rj, _market_history_econ_items_map, _market_history_listings_map)

        return self._parse_history_events(rj, _market_history_listings_map), rj["total_count"]

    async def my_market_history(
        self,
        *,
        start: int = 0,
        count: int = 100,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
        _market_history_econ_items_map: dict[str, MarketHistoryListingItem] = None,
        _market_history_listings_map: dict[int, MarketHistoryListing] = None,
    ) -> AsyncIterator[MY_MARKET_HISTORY_DATA]:
        """
        Fetch market history of self. Return async iterator to paginate over history event pages.

        :param start: start index
        :param count: listings per page. Steam do not accept value greater than 100
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields list of `MarketHistoryEvent`, total_count
        :raises EResultError: for ordinary reasons
        :raises SessionExpired:
        """

        if _item_descriptions_map is not None:
            _item_descriptions_map = {}
        if _market_history_econ_items_map is not None:
            _market_history_econ_items_map = {}
        if _market_history_listings_map is not None:
            _market_history_listings_map = {}

        more_listings = True
        while more_listings:
            # avoid excess destructuring
            history_data = await self.get_my_market_history(
                start=start,
                count=count,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
                _market_history_econ_items_map=_market_history_econ_items_map,
                _market_history_listings_map=_market_history_listings_map,
            )
            start += count
            more_listings = history_data[1] > start

            yield history_data

    @staticmethod
    def _parse_assets_for_history_listings(
        data: dict[str, dict[str, dict[str, dict]]],
        item_descrs_map: T_SHARED_DESCRIPTIONS,
        econ_item_map: dict[str, MarketHistoryListingItem],
    ):
        for app_id, app_data in data.items():
            for context_id, context_data in app_data.items():
                for a_data in context_data.values():
                    # because I don't know why in data `id` and `unowned_id` combinations and how that suppose to work
                    key_id = create_ident_code(a_data["id"], context_id, app_id)
                    key_unowned_id = create_ident_code(a_data["unowned_id"], context_id, app_id)
                    if key_id not in item_descrs_map or key_unowned_id not in item_descrs_map:
                        econ_item = MarketHistoryListingItem(
                            asset_id=int(a_data["id"]),
                            unowned_id=int(a_data["unowned_id"]),
                            unowned_context_id=int(a_data["unowned_contextid"]),
                            rollback_new_asset_id=int(a_data["rollback_new_id"])
                            if "rollback_new_id" in a_data
                            else None,
                            rollback_new_context_id=int(a_data["rollback_new_contextid"])
                            if "rollback_new_contextid" in a_data
                            else None,
                            app_context=AppContext((App(int(a_data["appid"])), int(a_data["contextid"]))),
                            description=item_descrs_map[
                                create_ident_code(
                                    a_data["instanceid"],
                                    a_data["classid"],
                                    app_id,
                                )
                            ],
                        )
                        if key_id not in econ_item_map:
                            econ_item_map[key_id] = econ_item
                        if key_unowned_id not in econ_item_map:
                            econ_item_map[key_unowned_id] = econ_item

    @staticmethod
    def _parse_history_listings(
        data: dict[str, dict[str, dict]],
        econ_item_map: dict[str, MarketHistoryListingItem],
        listings_map: dict[str, MarketHistoryListing],
    ):
        for l_id, l_data in data["listings"].items():  # sell listings
            if l_id not in listings_map:
                listings_map[l_id] = MarketHistoryListing(
                    id=int(l_data["listingid"]),
                    currency=Currency(int(l_data["currencyid"]) - 2000),
                    price=int(l_data["price"]),
                    fee=int(l_data["fee"]),
                    item=econ_item_map[
                        create_ident_code(
                            l_data["asset"]["id"],
                            l_data["asset"]["contextid"],
                            l_data["asset"]["appid"],
                        )
                    ],
                    original_price=int(l_data["original_price"]),
                    cancel_reason=l_data.get("cancel_reason"),
                )

        for p_id, p_data in data["purchases"].items():  # purchases :)
            if p_id not in listings_map:
                listing = MarketHistoryListing(
                    id=int(p_data["listingid"]),
                    currency=Currency(int(p_data["currencyid"]) - 2000),
                    received_currency=Currency(int(p_data["received_currencyid"]) - 2000),
                    paid_fee=int(p_data["paid_fee"]),
                    steam_fee=int(p_data["steam_fee"]),
                    publisher_fee=int(p_data["publisher_fee"]),
                    time_sold=datetime.fromtimestamp(p_data["time_sold"]),
                    item=econ_item_map[
                        create_ident_code(
                            p_data["asset"]["id"],
                            p_data["asset"]["contextid"],
                            p_data["asset"]["appid"],
                        )
                    ],
                    purchase_id=int(p_data["purchaseid"]),
                    steamid_purchaser=int(p_data["steamid_purchaser"]),
                    received_amount=int(p_data["received_amount"]),
                )
                listing.item.new_asset_id = int(p_data["asset"]["new_id"])
                listing.item.new_context_id = int(p_data["asset"]["new_contextid"])

                listings_map[p_id] = listing

    @staticmethod
    def _parse_history_events(
        data: dict[str, list[dict] | dict[str, dict]],
        listings_map: dict[str, MarketHistoryListing],
    ) -> list[MarketHistoryEvent]:
        events = []
        for e_data in data["events"]:
            listing_key = e_data["listingid"]
            if "purchaseid" in e_data:
                listing_key += "_" + e_data["purchaseid"]

            events.append(
                MarketHistoryEvent(
                    listing=listings_map[listing_key],
                    time_event=datetime.fromtimestamp(e_data["time_event"]),
                    type=MarketHistoryEventType(e_data["event_type"]),
                )
            )

        return events

    @overload
    async def fetch_price_history(
        self,
        obj: ItemDescription,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> list[PriceHistoryEntry]:
        ...

    @overload
    async def fetch_price_history(
        self,
        obj: str,
        app: App,
        *,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> list[PriceHistoryEntry]:
        ...

    async def fetch_price_history(
        self,
        obj: str,
        app: App = None,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
    ) -> list[PriceHistoryEntry]:
        """
        Fetch price history.
        Prices always will be same currency as a wallet.

        .. seealso:: https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Price-History

        .. note:: This request is rate limited by Steam.

        :param obj: `ItemDescription` or market hash name
        :param app: `Steam` app
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `PriceHistoryEntry`
        :raises EResultError:
        """

        if isinstance(obj, ItemDescription):
            name = obj.market_hash_name
            app = obj.app
        else:  # str
            name = obj

        params = {"appid": app.value, "market_hash_name": name, **params}
        r = await self.session.get(STEAM_URL.MARKET / "pricehistory", params=params, headers=headers)
        rj: dict[str, list[list]] = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError(rj.get("message", "Failed to fetch price history"), success, rj)

        return [
            PriceHistoryEntry(
                date=datetime.strptime(e_data[0], "%b %d %Y"),
                price=e_data[1],
                daily_volume=int(e_data[2]),
            )
            for e_data in rj["prices"]
        ]

    async def get_market_availability_info(self) -> tuple[bool, datetime | None]:
        """
        Return market status of whether it is available or not and, if possible, "when" the market will be available

        .. seealso:: https://help.steampowered.com/en/faqs/view/451E-96B3-D194-50FC
        """

        r = await self.session.get(STEAM_URL.MARKET, headers={"Referer": str(STEAM_URL.STORE)})
        rt = await r.text()

        available = True
        when: datetime | None = None

        if "The Market is unavailable for the following reason" in rt:
            available = False

            if "dateCanUseMarket" in rt:
                when = datetime.strptime(
                    re_search(r"var dateCanUseMarket = new Date\(\"(?P<date>.+)\"\);", rt).group("date"),
                    "%a, %d %b %Y %H:%M:%S %z",
                )

        # it would be nice if we can retrieve purchase history from https://store.steampowered.com/account/history/
        # domain (not steamcommunity, unfortunately, so cookies are separated), parse it and calc estimate market block
        # datetime, depends on last store purchase

        return available, when

    async def is_market_available(self) -> bool:
        info = await self.get_market_availability_info()
        return info[0]



from warnings import warn
from inspect import signature

from yarl import URL
from aiohttp import ClientSession, InvalidURL

try:
    from aiohttp_socks import ProxyConnector
except ImportError:
    ProxyConnector = None

from ..constants import STEAM_URL, Language
from ..utils import (
    get_cookie_value_from_session,
    remove_cookie_from_session,
    patch_session_with_http_proxy,
    add_cookie_to_session,
)

SESSION_ID_COOKIE = "sessionid"
LANG_COOKIE = "Steam_Language"
TZ_OFFSET_COOKIE = "timezoneOffset"
COOKIE_URLS = (STEAM_URL.COMMUNITY, STEAM_URL.STORE, STEAM_URL.HELP)


class SteamHTTPTransportMixin:
    """Handler of session instance, proxy, helper cookies getters/setters."""

    # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/transports/WebApiTransport.ts

    __slots__ = ()

    # required instance attributes
    session: ClientSession  # to use proxy session need to be patched

    @property
    def user_agent(self) -> str | None:
        return self.session.headers.get("User-Agent")

    @user_agent.setter
    def user_agent(self, value: str | None):
        if value is None:
            self.session.headers.pop("User-Agent", None)
        else:
            self.session.headers["User-Agent"] = value

    @property
    def language(self) -> Language:
        """Language of Steam html pages, json info, descriptions, etc."""
        return Language(get_cookie_value_from_session(self.session, STEAM_URL.COMMUNITY, LANG_COOKIE))

    @language.setter
    def language(self, value: Language | None):
        for url in COOKIE_URLS:
            if value is None:
                remove_cookie_from_session(self.session, url, LANG_COOKIE)
            else:
                add_cookie_to_session(self.session, url, LANG_COOKIE, value.value, secure=True)

    @property
    def tz_offset(self) -> str:
        return get_cookie_value_from_session(self.session, STEAM_URL.COMMUNITY, TZ_OFFSET_COOKIE)

    @tz_offset.setter
    def tz_offset(self, value: str | None):
        for url in COOKIE_URLS:
            if value is None:
                remove_cookie_from_session(self.session, url, TZ_OFFSET_COOKIE)
            else:
                add_cookie_to_session(self.session, url, TZ_OFFSET_COOKIE, value, samesite=True)

    # because this cookie set to guests also
    @property
    def session_id(self) -> str | None:
        """`sessionid` cookie value for `Steam Community` domain (https://steamcommunity.com)"""
        return self.get_session_id(STEAM_URL.COMMUNITY)

    @session_id.setter
    def session_id(self, value: str | None):
        self.set_session_id(STEAM_URL.COMMUNITY, value)

    # https://github.com/DoctorMcKay/node-steamcommunity/blob/7c564c1453a5ac413d9312b8cf8fe86e7578b309/index.js#L177
    def get_session_id(self, domain: URL) -> str | None:
        """Get `sessionid` cookie value for `Steam` domain"""

        return get_cookie_value_from_session(self.session, domain, SESSION_ID_COOKIE)

    def set_session_id(self, domain: URL, value: str | None):
        """Set `sessionid` cookie value for `Steam` domain"""

        if value is None:
            remove_cookie_from_session(self.session, domain, SESSION_ID_COOKIE)
        else:
            add_cookie_to_session(self.session, domain, SESSION_ID_COOKIE, value, samesite="None", secure=True)

    @staticmethod
    def _session_helper(session: ClientSession = None, proxy: str = None) -> ClientSession:
        """
        Helper function. Creates new `ClientSession` instance, patch/bound it to proxy if needed.
        Check passed session for `raise_for_status`.
        """

        if proxy and session:
            raise ValueError("You need to handle proxy connection by yourself with predefined session instance")
        elif proxy:
            if "socks" in proxy:
                if ProxyConnector is None:
                    raise TypeError(
                        """
                        To use `socks` type proxies you need `aiohttp_socks` package. 
                        You can do this with `aiosteampy[socks]` or `aiosteampy[all]` dependency install targets.
                        """
                    )

                # let aiohttp_socks parse url by herself
                session = ClientSession(connector=ProxyConnector.from_url(proxy), raise_for_status=True)
            else:  # http/s
                try:
                    proxy = URL(proxy)
                except ValueError as e:
                    raise InvalidURL(proxy) from e

                session = patch_session_with_http_proxy(ClientSession(raise_for_status=True), proxy)

        elif session:
            if not session._raise_for_status:
                warn(
                    "A session instance must be created with `raise_for_status=True` for client to work properly",
                    category=UserWarning,
                )
        else:  # nothing passed
            session = ClientSession(raise_for_status=True)

        return session

    # _proxy attr will be much easier, straight and less error-prone, why I need this?
    @property
    def proxy(self) -> str | None:
        """Proxy url in format `scheme://username:password@host:port` or `scheme://host:port`"""

        if isinstance(self.session.connector, ProxyConnector or int):  # socks, int to avoid TypeError if None
            c: ProxyConnector = self.session.connector

            scheme = str(c._proxy_type.name).lower()
            username = c._proxy_username
            password = c._proxy_password
            host = c._proxy_host
            port = c._proxy_port

        else:
            def_arg: URL | None = signature(self.session._request).parameters["proxy"].default  # magic

            if def_arg is None:  # client without proxy
                return None

            scheme = def_arg.scheme
            username = def_arg.user
            password = def_arg.password
            host = def_arg.host
            port = def_arg.port

        if username and password:  # with auth
            return f"{scheme}://{username}:{password}@{host}:{port}"
        else:
            return f"{scheme}://{host}:{port}"



"""
Helper functions to restore/load, dump SteamClient client and components,
decorators to check instance attributes.
"""

from typing import TYPE_CHECKING

try:
    from aiohttp_socks import ProxyConnector
except ImportError:
    ProxyConnector = None

from .utils import JSONABLE_COOKIE_JAR, update_session_cookies, attribute_required

if TYPE_CHECKING:
    from .client import SteamClientBase

__all__ = (
    "restore_from_cookies",
    "currency_required",
    "identity_secret_required",
)


async def restore_from_cookies(cookies: JSONABLE_COOKIE_JAR, client: "SteamClientBase") -> bool:
    """
    Helper func. Restore client session from cookies. Login if session is not alive.
    Return `True` if cookies are valid and not expired.
    """

    update_session_cookies(client.session, cookies)

    if not (await client.is_session_alive()):  # session initiated here
        await client.login(init_session=False)
        return False
    else:
        return True


# TODO restore from object/dict, dump to object/dict


currency_required = attribute_required(
    "currency",
    "You must provide a currency to client or init data before use this method",
)

identity_secret_required = attribute_required(
    "_identity_secret",
    "You must provide identity secret to client before use this method",
)



from typing import overload, Literal, TypeAlias, Sequence, AsyncIterator
from datetime import datetime
from json import dumps as jdumps

from yarl import URL

from ..typed import TradeOffersSummary
from ..constants import STEAM_URL, CORO, T_HEADERS, T_PARAMS, T_PAYLOAD, AppContext, App
from ..exceptions import EResultError, SteamError
from ..models import (
    TradeOffer,
    TradeOfferItem,
    TradeOfferStatus,
    HistoryTradeOffer,
    HistoryTradeOfferItem,
    EconItem,
)
from ..utils import create_ident_code, to_int_boolean, steam_id_to_account_id, account_id_to_steam_id
from .public import SteamCommunityPublicMixin, T_SHARED_DESCRIPTIONS
from .web_api import SteamWebApiMixin

T_TRADE_OFFERS_DATA: TypeAlias = tuple[list[TradeOffer], list[TradeOffer], int]


class TradeMixin(SteamWebApiMixin, SteamCommunityPublicMixin):
    """
    Mixin with trade offers related methods.
    Depends on `ConfirmationMixin`, `SteamCommunityPublicMixin`.
    """

    __slots__ = ()

    async def get_trade_offer(
        self,
        offer_id: int,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> TradeOffer:
        """
        Fetch trade offer from Steam.

        :param offer_id:
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :raises EResultError: for ordinary reasons
        """

        params = {
            "tradeofferid": offer_id,
            "language": self.language,
            "get_descriptions": 1,
            **params,
        }

        try:
            rj = await self.call_web_api(STEAM_URL.API.IEconService.GetTradeOffer, params=params, headers=headers)
        except EResultError as e:
            raise EResultError("Failed to fetch trade offer", e.result, e.data) from e

        data: dict[str, dict | list[dict]] = rj["response"]

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        self._update_item_descrs_map_from_trades(data["descriptions"], _item_descriptions_map)

        return self._create_trade_offer(data["offer"], _item_descriptions_map)

    @classmethod
    def _update_item_descrs_map_from_trades(cls, descrs: list[dict], item_descrs_map: T_SHARED_DESCRIPTIONS):
        for d_data in descrs:
            key = create_ident_code(d_data["instanceid"], d_data["classid"], d_data["appid"])
            if key not in item_descrs_map:
                item_descrs_map[key] = cls._create_item_descr(d_data)

    def _create_trade_offer(self, data: dict, item_descrs_map: T_SHARED_DESCRIPTIONS) -> TradeOffer:
        return TradeOffer(
            id=int(data["tradeofferid"]),
            owner_id=self.steam_id,
            partner_id=data["accountid_other"],
            is_our_offer=data["is_our_offer"],
            expiration_time=datetime.fromtimestamp(data["expiration_time"]),
            time_created=datetime.fromtimestamp(data["time_created"]),
            time_updated=datetime.fromtimestamp(data["time_updated"]),
            items_to_give=self._parse_items_for_trade(data.get("items_to_give", ()), item_descrs_map),
            items_to_receive=self._parse_items_for_trade(data.get("items_to_receive", ()), item_descrs_map),
            message=data["message"],
            status=TradeOfferStatus(data["trade_offer_state"]),
        )

    @classmethod
    def _parse_items_for_trade(cls, items: list[dict], item_descrs_map: T_SHARED_DESCRIPTIONS) -> list[TradeOfferItem]:
        return [
            TradeOfferItem(
                asset_id=a_data["assetid"],
                amount=int(a_data["amount"]),
                missing=a_data["missing"],
                est_usd=int(a_data.get("est_usd", 0)),
                app_context=AppContext((App(int(a_data["appid"])), int(a_data["contextid"]))),
                description=item_descrs_map.get(
                    create_ident_code(
                        a_data["instanceid"],
                        a_data["classid"],
                        a_data["appid"],
                    )
                ),
            )
            for a_data in items
        ]

    @overload
    async def get_trade_offers(
        self,
        *,
        time_historical_cutoff: int | datetime = ...,
        sent: bool = ...,
        received: bool = ...,
        cursor: int = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> T_TRADE_OFFERS_DATA:
        ...

    @overload
    async def get_trade_offers(
        self,
        *,
        historical_only: Literal[True] = ...,
        sent: bool = ...,
        received: bool = ...,
        cursor: int = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> T_TRADE_OFFERS_DATA:
        ...

    async def get_trade_offers(
        self,
        *,
        active_only=True,
        time_historical_cutoff: int | datetime = None,
        historical_only=False,
        sent=True,
        received=True,
        cursor=0,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> T_TRADE_OFFERS_DATA:
        """
        Fetch trade offers from `Steam Web Api`.

        .. note:: You can paginate by yourself passing `cursor` arg.
            Returned cursor with 0 value means that there is no more pages

        .. seealso:: https://steamapi.xpaw.me/#IEconService/GetTradeOffers

        :param active_only: fetch active, changed since `time_historical_cutoff` tradeoffs only or not
        :param time_historical_cutoff: timestamp for `active_only`
        :param historical_only: opposite for `active_only`
        :param sent: include sent offers or not
        :param received: include received offers or not
        :param cursor: cursor integer, need to paginate over
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: sent trades, received trades lists, next cursor
        :raises EResultError: for ordinary reasons
        :raises TypeError:
        """

        if historical_only:
            if time_historical_cutoff is not None:
                raise TypeError("Argument `time_historical_cutoff` must be used only with `active_only`")

            active_only = False

        params = {
            "active_only": to_int_boolean(active_only),
            "get_sent_offers": to_int_boolean(sent),
            "get_received_offers": to_int_boolean(received),
            "historical_only": to_int_boolean(historical_only),
            "get_descriptions": 1,
            "cursor": cursor,
            "language": self.language,
            **params,
        }
        if active_only and time_historical_cutoff is not None:
            if isinstance(time_historical_cutoff, datetime):
                params["time_historical_cutoff"] = int(time_historical_cutoff.timestamp())  # trunc ms
            else:  # int ts
                params["time_historical_cutoff"] = time_historical_cutoff

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        try:
            rj = await self.call_web_api(STEAM_URL.API.IEconService.GetTradeOffers, params=params, headers=headers)
        except EResultError as e:
            raise EResultError("Failed to fetch trade offers", e.result, e.data) from e

        data: dict[str, dict | list[dict]] = rj["response"]

        self._update_item_descrs_map_from_trades(data["descriptions"], _item_descriptions_map)

        return (
            [self._create_trade_offer(d, _item_descriptions_map) for d in data.get("trade_offers_sent", ())],
            [self._create_trade_offer(d, _item_descriptions_map) for d in data.get("trade_offers_received", ())],
            data.get("next_cursor", 0),
        )

    @overload
    async def trade_offers(
        self,
        *,
        time_historical_cutoff: int | datetime = ...,
        sent: bool = ...,
        received: bool = ...,
        cursor: int = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> AsyncIterator[T_TRADE_OFFERS_DATA]:
        ...

    @overload
    async def trade_offers(
        self,
        *,
        historical_only: Literal[True] = ...,
        sent: bool = ...,
        received: bool = ...,
        cursor: int = ...,
        params: T_PARAMS = ...,
        headers: T_HEADERS = ...,
    ) -> AsyncIterator[T_TRADE_OFFERS_DATA]:
        ...

    async def trade_offers(
        self,
        *,
        active_only=True,
        time_historical_cutoff: int | datetime = None,
        historical_only=False,
        sent=True,
        received=True,
        cursor=0,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> AsyncIterator[T_TRADE_OFFERS_DATA]:
        """
        Fetch trade offers from `Steam Web Api`. Return async iterator to paginate over offers pages.

        .. note:: Method requires API key

        :param active_only: fetch active, changed since `time_historical_cutoff` tradeoffs only or not
        :param time_historical_cutoff: timestamp for `active_only`
        :param historical_only: opposite for `active_only`
        :param sent: include sent offers or not
        :param received: include received offers or not
        :param cursor: cursor integer, need to paginate over offer pages
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `AsyncIterator` that yields sent trades, received trades lists, next cursor
        :raises EResultError: for ordinary reasons
        :raises TypeError:
        """

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        more_offers = True
        while more_offers:
            # avoid excess destructuring
            offers_data = await self.get_trade_offers(
                active_only=active_only,
                time_historical_cutoff=time_historical_cutoff,
                historical_only=historical_only,
                sent=sent,
                received=received,
                cursor=cursor,
                params=params,
                headers=headers,
                _item_descriptions_map=_item_descriptions_map,
            )
            cursor = offers_data[2]
            more_offers = bool(cursor)

            yield offers_data

    async def get_trade_offers_summary(self, *, params: T_PARAMS = {}, headers: T_HEADERS = {}) -> TradeOffersSummary:
        """
        Get trade offers summary from Steam Web Api.

        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: trade offers summary
        :raises EResultError:
        """

        try:
            rj: dict[str, TradeOffersSummary] = await self.call_web_api(
                STEAM_URL.API.IEconService.GetTradeOffersSummary,
                params=params,
                headers=headers,
            )
        except EResultError as e:
            raise EResultError("Failed to fetch trade offers summary", e.result, e.data) from e

        return rj["response"]

    async def get_trade_receipt(
        self,
        offer_id: int,
        *,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> HistoryTradeOffer:
        """
        Fetch single trade offer from history.

        :param offer_id:
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: `HistoryTradeOffer`
        :raises EResultError: for ordinary reasons
        """

        params = {
            "tradeid": offer_id,
            "get_descriptions": 1,
            "language": self.language,
            **params,
        }
        try:
            rj = await self.call_web_api(STEAM_URL.API.IEconService.GetTradeStatus, params=params, headers=headers)
        except EResultError as e:
            raise EResultError("Failed to fetch trade receipt", e.result, e.data) from e

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        data: dict[str, int | bool | dict | list[dict]] = rj["response"]
        self._update_item_descrs_map_from_trades(data["descriptions"], _item_descriptions_map)

        return self._create_history_trade_offer(data["trades"][0], _item_descriptions_map)

    async def get_trade_history(
        self,
        *,
        max_trades=100,
        start_after_time=0,
        start_after_trade_id=0,
        navigating_back=False,
        include_failed=True,
        params: T_PARAMS = {},
        headers: T_HEADERS = {},
        _item_descriptions_map: T_SHARED_DESCRIPTIONS = None,
    ) -> tuple[list[HistoryTradeOffer], int]:
        """
        Fetch history trades with changed assets data.
        You can paginate by yourself with this method.

        .. note:: Method requires API key

        :param max_trades: page size
        :param start_after_time: timestamp
        :param start_after_trade_id:
        :param navigating_back:
        :param include_failed:
        :param params: extra params to pass to url
        :param headers: extra headers to send with request
        :return: list of `HistoryTradeOffer`, total trades count
        :raises EResultError: for ordinary reasons
        """

        params = {
            "max_trades": max_trades,
            "get_descriptions": 1,
            "include_total": 1,
            "include_failed": to_int_boolean(include_failed),
            "navigating_back": to_int_boolean(navigating_back),
            "start_after_time": start_after_time,
            "start_after_tradeid": start_after_trade_id,
            "language": self.language,
            **params,
        }
        try:
            rj = await self.call_web_api(STEAM_URL.API.IEconService.GetTradeHistory, params=params, headers=headers)
        except EResultError as e:
            raise EResultError("Failed to fetch trades history", e.result, e.data) from e

        if _item_descriptions_map is None:
            _item_descriptions_map = {}

        data: dict[str, int | bool | dict | list[dict]] = rj["response"]
        self._update_item_descrs_map_from_trades(data["descriptions"], _item_descriptions_map)

        return (
            [self._create_history_trade_offer(d, _item_descriptions_map) for d in data["trades"]],
            data["total_trades"],
        )

    def _create_history_trade_offer(self, data: dict, item_descrs_map: T_SHARED_DESCRIPTIONS) -> HistoryTradeOffer:
        return HistoryTradeOffer(
            id=int(data["tradeid"]),
            owner_id=self.steam_id,
            partner_id=steam_id_to_account_id(int(data["steamid_other"])),
            time_init=datetime.fromtimestamp(data["time_init"]),
            status=TradeOfferStatus(data["status"]),
            assets_given=self._parse_items_for_history_trades(data.get("assets_given", ()), item_descrs_map),
            assets_received=self._parse_items_for_history_trades(data.get("assets_received", ()), item_descrs_map),
        )

    @classmethod
    def _parse_items_for_history_trades(
        cls,
        items: list[dict],
        item_descrs_map: T_SHARED_DESCRIPTIONS,
    ) -> list[HistoryTradeOfferItem]:
        return [
            HistoryTradeOfferItem(
                asset_id=int(a_data["assetid"]),
                amount=int(a_data["amount"]),
                new_asset_id=int(a_data["new_assetid"]),
                new_context_id=int(a_data["new_contextid"]),
                app_context=AppContext((App(int(a_data["appid"])), int(a_data["contextid"]))),
                description=item_descrs_map.get(
                    create_ident_code(
                        a_data["instanceid"],
                        a_data["classid"],
                        a_data["appid"],
                    )
                ),
            )
            for a_data in items
        ]

    async def perform_action_with_offer(
        self,
        offer_id: int,
        action: Literal["cancel", "decline"],
        payload: T_PAYLOAD,
        headers: T_HEADERS,
    ) -> dict[str, str]:
        """Cancel or decline trade offer"""

        r = await self.session.post(
            STEAM_URL.TRADE / f"{offer_id}/{action}",
            data={"sessionid": self.session_id, **payload},
            headers=headers,
        )
        # TODO TypedDict
        return await r.json()

    async def cancel_trade_offer(self, obj: int | TradeOffer, *, payload: T_PAYLOAD = {}, headers: T_HEADERS = {}):
        """
        Cancel outgoing trade offer.

        :param obj:
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :raises TypeError:
        :raises EResultError:
        """

        if isinstance(obj, TradeOffer):
            if not obj.is_our_offer:
                raise TypeError("You can't cancel not your offer! Are you trying to decline incoming offer?")
            offer_id = obj.id
        else:
            offer_id = obj
        offer_data = await self.perform_action_with_offer(offer_id, "cancel", payload, headers)
        resp_offer_id = int(offer_data.get("tradeofferid", 0))
        if not resp_offer_id or resp_offer_id != offer_id:
            raise SteamError(f"Failed to cancel trade offer", offer_data)

    async def decline_trade_offer(self, obj: int | TradeOffer, *, payload: T_PAYLOAD = {}, headers: T_HEADERS = {}):
        """
        Decline incoming trade offer.

        :param obj:
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :raises TypeError:
        :raises EResultError:
        """

        if isinstance(obj, TradeOffer):
            if obj.is_our_offer:
                raise TypeError("You can't decline your offer! Are you trying to cancel outgoing offer?")
            offer_id = obj.id
        else:
            offer_id = obj
        offer_data = await self.perform_action_with_offer(offer_id, "decline", payload, headers)
        resp_offer_id = int(offer_data.get("tradeofferid", 0))
        if not resp_offer_id or resp_offer_id != offer_id:
            raise SteamError(f"Failed to decline trade offer", offer_data)

    @overload
    async def accept_trade_offer(
        self,
        obj: TradeOffer,
        *,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ):
        ...

    @overload
    async def accept_trade_offer(
        self,
        obj: int,
        partner: int = ...,
        *,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ):
        ...

    async def accept_trade_offer(
        self,
        obj: int | TradeOffer,
        partner: int = None,
        *,
        confirm=True,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
    ):
        """
        Accept trade offer, yes.

        .. note::
            Auto confirm accepting if needed.

            If you not pass `partner` but pass `trade offer id` -
            fetches :class:`aiosteampy.models.TradeOffer` from `Steam`.

        :param obj: `TradeOffer` or trade offer id
        :param partner: partner account id (id32) or steam id (id64)
        :param confirm: auto-confirm offer
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :raises EResultError: if there is error when trying to fetch `TradeOffer`
        :raises TypeError:
        """

        if isinstance(obj, TradeOffer):
            if obj.is_our_offer:
                raise TypeError("You can't accept your offer! Are you trying to cancel outgoing offer?")
            offer_id = obj.id
            partner = obj.partner_id64
        else:  # int
            if not partner:
                fetched = await self.get_trade_offer(obj)
                return await self.accept_trade_offer(fetched)

            offer_id = obj
            partner = account_id_to_steam_id(partner) if partner < 4294967296 else partner  # 2**32

        data = {
            "sessionid": self.session_id,
            "tradeofferid": offer_id,
            "serverid": 1,
            "partner": partner,
            "captcha": "",
            **payload,
        }
        url_base = STEAM_URL.TRADE / str(offer_id)
        r = await self.session.post(url_base / "accept", data=data, headers={"Referer": str(url_base), **headers})
        rj: dict = await r.json()
        if rj.get("needs_mobile_confirmation") and confirm:
            await self.confirm_trade_offer(offer_id)

    @overload
    async def make_trade_offer(
        self,
        obj: int,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message: str = ...,
        *,
        token: str = ...,
        confirm: bool = ...,
        countered_id: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def make_trade_offer(
        self,
        obj: str,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message: str = ...,
        *,
        confirm: bool = ...,
        countered_id: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def make_trade_offer(
        self,
        obj: int,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message: str = ...,
        *,
        token: str = ...,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        countered_id: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> TradeOffer:
        ...

    @overload
    async def make_trade_offer(
        self,
        obj: str,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message: str = ...,
        *,
        fetch: Literal[True] = ...,
        confirm: bool = ...,
        countered_id: int = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> TradeOffer:
        ...

    async def make_trade_offer(
        self,
        obj: int | str,
        to_give: Sequence[EconItem] = (),
        to_receive: Sequence[EconItem] = (),
        message="",
        *,
        token: str = None,
        fetch=False,
        confirm=True,
        countered_id: int = None,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
    ) -> int | TradeOffer:
        """
        Make (send) steam trade offer to partner.

        .. note:: Make sure that partner is in friends list if you not pass trade url or trade token.

        :param obj: partner trade url, partner id(id32 or id64)
        :param token: trade token (mandatory if `obj` is partner id)
        :param fetch: make a request, fetch and return trade offer
        :param to_give: sequence of items that you want to give
        :param to_receive: sequence of items that you want to receive
        :param message: message to the partner
        :param confirm: auto-confirm offer
        :param countered_id: id of offer that you want to counter. Use `counter_trade_offer` method for this
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :return: trade offer id
        :raises ValueError: trade is empty
        """

        if not to_give and not to_receive:
            raise ValueError("You can't make empty trade offer!")

        trade_url, partner_id32, partner_id64, token = self._parse_make_offer_args(obj, token)
        base_url = STEAM_URL.TRADE / "new/"
        referer = trade_url or base_url % {"partner": partner_id32}
        offer_params = {}
        if token:
            referer %= {"token": token}
            offer_params["trade_offer_access_token"] = token
        data = {
            "sessionid": self.session_id,
            "serverid": 1,
            "partner": partner_id64,
            "tradeoffermessage": message,
            "json_tradeoffer": jdumps(
                {
                    "newversion": True,
                    "version": len(to_give) + len(to_receive) + 1,
                    "me": {"assets": [self._to_asset_dict(i) for i in to_give], "currency": [], "ready": False},
                    "them": {"assets": [self._to_asset_dict(i) for i in to_receive], "currency": [], "ready": False},
                }
            ),
            "captcha": "",
            "trade_offer_create_params": jdumps(offer_params),
            **payload,
        }
        if countered_id:
            data["tradeofferid_countered"] = countered_id

        r = await self.session.post(base_url / "send", data=data, headers={"Referer": str(referer), **headers})
        rj = await r.json()
        offer_id = int(rj["tradeofferid"])
        if confirm and rj.get("needs_mobile_confirmation"):
            conf = await self.confirm_trade_offer(offer_id)
            offer_id = conf.creator_id

        return await self.get_trade_offer(offer_id, headers=headers) if fetch else offer_id

    @staticmethod
    def _to_asset_dict(i: EconItem) -> dict[str, int | str]:
        return {
            "appid": i.app_context.app.value,
            "contextid": str(i.app_context.context),
            "amount": i.amount,
            "assetid": str(i.asset_id),
        }

    @staticmethod
    def _parse_make_offer_args(obj: str | int, token: str | None) -> tuple[str | None, int, int, str | None]:
        trade_url = None
        if isinstance(obj, str):  # trade url
            trade_url = URL(obj)
            partner_id32 = int(trade_url.query["partner"])
            partner_id64 = account_id_to_steam_id(partner_id32)
            token = trade_url.query["token"]
            trade_url = str(trade_url)
        else:
            partner_id = obj
            if partner_id < 4294967296:  # 32
                partner_id32 = partner_id
                partner_id64 = account_id_to_steam_id(partner_id)
            else:  # 64
                partner_id64 = partner_id
                partner_id32 = steam_id_to_account_id(partner_id64)

        return trade_url, partner_id32, partner_id64, token

    @overload
    async def counter_trade_offer(
        self,
        obj: TradeOffer,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message="",
        *,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    @overload
    async def counter_trade_offer(
        self,
        obj: int,
        to_give: Sequence[EconItem] = ...,
        to_receive: Sequence[EconItem] = ...,
        message="",
        *,
        partner_id: int,
        confirm: bool = ...,
        payload: T_PAYLOAD = ...,
        headers: T_HEADERS = ...,
    ) -> int:
        ...

    def counter_trade_offer(
        self,
        obj: TradeOffer | int,
        to_give: Sequence[EconItem] = (),
        to_receive: Sequence[EconItem] = (),
        message="",
        *,
        partner_id: int = None,
        confirm=True,
        payload: T_PAYLOAD = {},
        headers: T_HEADERS = {},
    ) -> CORO[int]:
        """
        Counter trade offer with another.

        :param obj: `TradeOffer` or trade offer id of which you want to counter
        :param to_give: sequence of items that you want to give
        :param to_receive: sequence of items that you want to receive
        :param message: message to the partner
        :param partner_id:
        :param confirm: auto-confirm offer
        :param payload: extra payload data
        :param headers: extra headers to send with request
        :return: trade offer id
        :raises TypeError:
        """

        if isinstance(obj, TradeOffer):
            offer_id = obj.id
            to_give_updated = [*obj.items_to_give, *to_give]
            to_receive_updated = [*obj.items_to_receive, *to_receive]
            if obj.is_our_offer:
                raise TypeError("You can't counter your offer!")
        else:  # trade id
            offer_id = obj
            to_give_updated = to_give
            to_receive_updated = to_receive

        return self.make_trade_offer(
            partner_id,
            to_give_updated,
            to_receive_updated,
            message,
            confirm=confirm,
            countered_id=offer_id,
            payload=payload,
            headers=headers,
        )



import asyncio
from datetime import datetime, timedelta
from base64 import b64encode
from time import time as time_time

from yarl import URL
from aiohttp import ClientResponseError
from aiohttp.client import _RequestContextManager
from rsa import PublicKey, encrypt

from ..constants import STEAM_URL, EResult
from ..typed import JWTToken
from ..exceptions import LoginError, EResultError
from ..utils import (
    generate_session_id,
    get_cookie_value_from_session,
    remove_cookie_from_session,
    format_time,
    decode_jwt,
    add_cookie_to_session,
)
from .http import SESSION_ID_COOKIE
from .guard import SteamGuardMixin


REFERER_HEADER = {"Referer": str(STEAM_URL.COMMUNITY) + "/"}
# https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/helpers.ts#L17
API_HEADERS = {
    "accept": "application/json, text/plain, */*",
    "sec-fetch-site": "cross-site",
    "sec-fetch-mode": "cors",
    "sec-fetch-dest": "empty",
}
STEAM_SECURE_COOKIE = "steamLoginSecure"


class LoginMixin(SteamGuardMixin):
    """
    Mixin with login logic methods.
    Depends on `SteamGuardMixin`.
    """

    __slots__ = ()

    # required instance attributes
    username: str
    _password: str
    _refresh_token: str | None

    # better to cache it somehow
    @property
    def access_token(self) -> str | None:
        """
        Encoded `JWT access token` as cookie value for `Steam Community` domain (https://steamcommunity.com).
        Can be used to make requests to a `Steam Web API`
        """

        return self.get_access_token(STEAM_URL.COMMUNITY)

    # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L231
    @access_token.setter
    def access_token(self, token: str | None):
        if token is None:
            remove_cookie_from_session(self.session, STEAM_URL.COMMUNITY, STEAM_SECURE_COOKIE)
        else:
            # accept only community token as project intent to be a "steam community" representation
            # TODO checks
            # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L231

            add_cookie_to_session(
                self.session,
                STEAM_URL.COMMUNITY,
                STEAM_SECURE_COOKIE,
                token,
                expires=format_time(datetime.now() + timedelta(days=365 * 5)),
                samesite="None",
                secure=True,
                httponly=True,
            )

    @property
    def access_token_decoded(self) -> JWTToken | None:
        if token := self.access_token:
            return decode_jwt(token)

    @property
    def is_access_token_expired(self) -> bool:
        """If access token has expired. Also returns `True` if token is not yet set"""
        if token := self.access_token_decoded:
            return (token["exp"]) <= int(time_time())
        else:
            return True

    def get_access_token(self, domain: URL) -> str | None:
        """Get encoded `JWT access token` as cookie value for `Steam Domain`"""

        if token := get_cookie_value_from_session(self.session, domain, STEAM_SECURE_COOKIE):
            return token.split("%7C%7C")[1]  # ||

    @property
    def refresh_token(self) -> str | None:
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, token: str | None):
        # TODO checks
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L280
        self._refresh_token = token

    @property
    def refresh_token_decoded(self) -> JWTToken | None:
        if token := self._refresh_token:
            return decode_jwt(token)

    @property
    def is_refresh_token_expired(self) -> bool:
        """If refresh token has expired. Also returns `True` if token is not yet set"""
        if token := self.refresh_token_decoded:
            return (token["exp"]) <= int(time_time())
        else:
            return True

    async def is_session_alive(self, domain=STEAM_URL.COMMUNITY) -> bool:
        """Check if session is alive for `Steam` domain"""

        # we can also check https://steamcommunity.com/my for redirect to profile page as indicator
        # https://github.com/DoctorMcKay/node-steamcommunity/blob/1067d4572ee9d467e8f686951901c51028c5c995/index.js#L290

        # ensure that redirects is allowed and access token can be refreshed
        r = await self.session.get(domain, allow_redirects=True)
        rt = await r.text()
        return self.username in rt

    async def login(self, init_session=True):
        """
        Perform login for main `Steam` domains:
            * https://steamcommunity.com
            * https://store.steampowered.com
            * https://help.steampowered.com

        :param init_session: init session before start auth process.
            Set this to False if you already make requests to `Steam` from current client
        :raises EResultError: when failed to obtain rsa key, update steam guard code
        :raises LoginError: other login process errors
        """

        # https://github.com/bukson/steampy/blob/fe0433c8cf7020318cfbbc22e79028a7576374ee/steampy/login.py#L67
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/examples/login-to-web-with-2fa.ts#L13
        init_session and await self.session.get(STEAM_URL.COMMUNITY)

        session_data = await self._begin_auth_session_with_credentials()
        await self._update_auth_session_with_steam_guard_code(session_data)
        await self._poll_auth_session_status(session_data)
        fin_data = await self._finalize_login()  # there can be retrieved steam id

        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/LoginSession.ts#L845
        loop = asyncio.get_event_loop()
        transfers = [
            loop.create_task(self._perform_transfer(d, fin_data["steamID"])) for d in fin_data["transfer_info"]
        ]
        # there is no guarantee that first completed transfer will be to community and
        # steamLoginSecure cookie will be not present yet, so better to wait until all transfers completed,
        # and we can be sure that login process to community domain is done
        # moreover, steam domains (store, community, help, tv, login) has own access tokens
        await asyncio.wait(transfers, return_when=asyncio.ALL_COMPLETED)

        # Transfers exception check ?

    async def _perform_transfer(self, data: dict, steam_id: str | int = None):
        """Perform a transfer of params and tokens to steam login endpoints"""

        r = await self.session.post(data["url"], data={**data["params"], "steamID": steam_id or self.steam_id})
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L864
        # make sure that `steamLoginSecure` cookie is present
        if not r.cookies.get("steamLoginSecure"):
            raise LoginError("'steamLoginSecure' cookie is not present in result", r.cookies)

        # ensure that sessionid cookie is presented
        # https://github.com/DoctorMcKay/node-steam-session/blob/698469cdbad3e555dda10c81f580f1ee3960156f/src/LoginSession.ts#L872-L873
        if not r.cookies.get(SESSION_ID_COOKIE):
            add_cookie_to_session(
                self.session,
                r.real_url,
                SESSION_ID_COOKIE,
                generate_session_id(),
                samesite="None",
                secure=True,
            )

    async def _begin_auth_session_with_credentials(self) -> dict:
        pub_key, ts = await self._get_rsa_key()
        # for web browser
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/AuthenticationClient.ts#L390
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/enums-steam/EAuthTokenPlatformType.ts
        platform_data = {
            "website_id": "Community",
            "device_details": {
                "device_friendly_name": self.user_agent,
                "platform_type": 2,
            },
        }

        data = {
            "account_name": self.username,
            "encrypted_password": b64encode(encrypt(self._password.encode("utf-8"), pub_key)).decode(),
            "encryption_timestamp": ts,
            "remember_login": "true",
            "persistence": "1",
            **platform_data,
        }
        r = await self.session.post(
            STEAM_URL.API.IAuthService.BeginAuthSessionViaCredentials,
            data=data,
            headers=REFERER_HEADER,
        )
        return await r.json()

    async def _update_auth_session_with_steam_guard_code(self, session_data: dict):
        # Doesn't check allowed confirmations, but it's probably not needed
        # as steam accounts suited for trading must have a steam guard and device code.

        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/LoginSession.ts#L735
        # https://github.com/DoctorMcKay/node-steam-session/blob/64463d7468c1c860afb80164b8c5831e629f657f/src/enums-steam/EAuthSessionGuardType.ts
        data = {
            "client_id": session_data["response"]["client_id"],
            "steamid": session_data["response"]["steamid"],
            "code_type": 3,
            "code": self.two_factor_code,
        }

        try:
            await self.session.post(
                STEAM_URL.API.IAuthService.UpdateAuthSessionWithSteamGuardCode,
                data=data,
                headers=REFERER_HEADER,
            )
        except ClientResponseError as e:
            raise LoginError("Error updating steam guard code") from e

    async def _poll_auth_session_status(self, session_resp: dict):
        data = {
            "client_id": session_resp["response"]["client_id"],
            "request_id": session_resp["response"]["request_id"],
        }
        r = await self.session.post(
            STEAM_URL.API.IAuthService.PollAuthSessionStatus,
            data=data,
            headers=REFERER_HEADER,
        )
        rj = await r.json()
        if rj.get("response", {"had_remote_interaction": True})["had_remote_interaction"]:
            raise LoginError("Error polling auth session status", rj)

        self._refresh_token = rj["response"]["refresh_token"]
        # access token already presented in cookie, btw this token has "web" aud unlike others
        # self.access_token = rj["response"]["access_token"]

    async def _finalize_login(self) -> dict:
        data = {
            "nonce": self._refresh_token,
            "sessionid": self.session_id,
            "redir": str(STEAM_URL.COMMUNITY / "login/home/?goto="),
        }
        r = await self.session.post(
            STEAM_URL.LOGIN / "jwt/finalizelogin",
            data=data,
            headers={**API_HEADERS, **REFERER_HEADER, "Origin": str(STEAM_URL.COMMUNITY)},
        )
        rj: dict = await r.json()
        if rj and rj.get("error"):
            raise LoginError("Get error response when performing login finalization", rj)
        elif not rj or not rj.get("transfer_info"):
            raise LoginError("Malformed login response", rj)

        return rj

    async def _get_rsa_key(self) -> tuple[PublicKey, int]:
        r = await self.session.get(
            STEAM_URL.API.IAuthService.GetPasswordRSAPublicKey,
            params={"account_name": self.username},
        )
        rj = await r.json()
        try:
            rsa_mod = int(rj["response"]["publickey_mod"], 16)
            rsa_exp = int(rj["response"]["publickey_exp"], 16)
            rsa_timestamp = int(rj["response"]["timestamp"])

            return PublicKey(rsa_mod, rsa_exp), rsa_timestamp

        except KeyError:
            raise LoginError("Could not obtain rsa-key", rj)

    def logout(self) -> _RequestContextManager:
        return self.session.post(
            STEAM_URL.COMMUNITY / "login/logout/",
            data={**REFERER_HEADER, "sessionid": self.session_id},
            allow_redirects=False,
        )

    async def get_store_access_token_from_steam(self) -> str | None:
        """
        Fetch access token for `Steam Store` domain (https://store.steampowered.com/).
        Return token only if you logged in store domain.

        :return: raw token string or `None`
        :raises EResultError: for ordinary reasons
        """

        r = await self.session.get(STEAM_URL.STORE / "pointssummary/ajaxgetasyncconfig")
        rj = await r.json()
        success = EResult(rj.get("success"))
        if success is not EResult.OK:
            raise EResultError("Failed to fetch store access token", success, rj)

        return rj["data"]["webapi_token"] if rj["data"] else None

    async def refresh_access_token(self) -> str:
        """Request to refresh access token by web browser method"""

        await self.session.get(
            STEAM_URL.LOGIN / "jwt/refresh" % {"redir": str(STEAM_URL.COMMUNITY)},
            allow_redirects=True,
        )

        return self.access_token



from time import time as time_time

from ..helpers import identity_secret_required
from ..utils import (
    gen_two_factor_code,
    generate_confirmation_key,
    async_throttle,
    steam_id_to_account_id,
)
from .http import SteamHTTPTransportMixin


class SteamGuardMixin(SteamHTTPTransportMixin):
    """
    Mixin with Steam Guard related methods.
    Depends on `SteamHTTPTransportMixin`.
    """

    __slots__ = ()

    # required instance attributes
    steam_id: int
    device_id: str
    _shared_secret: str
    _identity_secret: str | None

    @property
    def account_id(self) -> int:
        """Steam id32."""
        return steam_id_to_account_id(self.steam_id)

    @property
    def two_factor_code(self) -> str:
        """Generate twofactor (onetime/TOTP) code."""
        return gen_two_factor_code(self._shared_secret)

    @async_throttle(1, arg_name="tag")
    @identity_secret_required
    async def _gen_confirmation_key(self, *, tag: str) -> tuple[str, int]:
        ts = int(time_time())
        return generate_confirmation_key(self._identity_secret, tag, ts), ts



"""Currency converter extension."""

import asyncio
from collections import UserDict
from datetime import datetime, timedelta

from yarl import URL
from aiohttp import ClientSession

try:
    from croniter import croniter

except ImportError:
    from warnings import warn

    warn(
        """
        The `aiosteampy.converter` module requires the `croniter` library to be installed
        to make the rates synchronization with backend service work.
        In order You need this functionality, You can use `aiosteampy[converter]` dependency install target.
        """,
        category=ImportWarning,
    )

    croniter = None


from ..constants import Currency

__all__ = ("CurrencyConverter", "API_URL", "SERT_CRON")

API_URL = URL("https://sert.somespecial.one/")
SERT_CRON = "9,39 * * * *"


class CurrencyConverter(UserDict[Currency, tuple[float, datetime]]):
    """
    Dict-like class to handle converting steam currencies.

    .. seealso:: https://github.com/somespecialone/sert
    """

    __slots__ = ("_cron", "api_url", "_sync_task")

    def __init__(self, *, api_url=API_URL, cron_exp=SERT_CRON):
        """
        :param api_url: url of `sert` backend service api
        :param cron_exp: cron expression string of `sert` backend service
        """

        super().__init__()

        self.api_url = api_url

        # normalize Deta cron expression
        self._cron = croniter(cron_exp, ret_type=datetime) if croniter else None
        self._sync_task: asyncio.Task | None = None

    @property
    def currencies(self) -> list[Currency]:
        return list(self.keys())

    @property
    def expired(self) -> bool:
        """If at least one single currency rate was updated yesterday at the latest."""

        utcnow_date = datetime.utcnow().date()
        return any(updated.date() != utcnow_date for _, updated in self.values())

    async def load(self):
        """Load rates."""

        async with ClientSession() as sess:
            r = await sess.get(self.api_url / "rates")
            rj: dict[str, list[float, int]] = await r.json()

        for curr_key, v in rj.items():
            self[getattr(Currency, curr_key)] = (v[0], datetime.utcfromtimestamp(v[1]))

    def close(self):
        """
        Cancel inner `_sync_task` if it exists.
        Needed if You use a synchronization task (call `synchronize`previously).
        """

        self._sync_task and self._sync_task.cancel()

    def synchronize(self):
        """
        Synchronize rates with backend.
        Create background task.

        If you want handle task by yourself you can use `get_wait_time` method
        to get time in seconds for the next rate update on backend.
        """

        async def sync_coro():
            while self._cron:
                await asyncio.gather(asyncio.sleep(self.get_wait_time()), self.load())

        self._sync_task = asyncio.create_task(sync_coro())

    def get_wait_time(self) -> float:
        """Calculate time to nearest service rates update."""

        utcnow = datetime.utcnow()
        if self.expired:
            when = self._cron.get_next(start_time=utcnow)
        else:
            # go to the next day
            when = self._cron.get_next(start_time=utcnow.replace(hour=0, minute=0, second=0) + timedelta(days=1))

        wait = (when - utcnow) + timedelta(minutes=1)  # offset to ensure that service finish to update rates
        return wait.total_seconds()

    def convert(self, amount: int, currency: Currency, target=Currency.USD) -> int:
        """
        Convert amount from `currency` to `target` currency.

        :param amount: amount of currency that need to be converted
        :param currency: passed currency of `amount`
        :param target: target currency of returned value
        :return: converted amount
        :raises KeyError: if provided currency is not present in `Converter`
        """

        source_rate = self[currency][0] if currency is not Currency.USD else 1
        target_rate = self[target][0] if target is not Currency.USD else 1

        # direct conversion
        # return round(amount * (target_rate / source_rate))

        # with USD in middle step
        usd_amount = round(amount * (1 / source_rate))
        return round(usd_amount * target_rate)



"""Constants and enums, some types"""

from typing import TypeAlias, Any, TypeVar, Coroutine, Mapping
from enum import Enum, IntEnum

from yarl import URL
from aenum import extend_enum

_T = TypeVar("_T")

CORO: TypeAlias = Coroutine[Any, Any, _T]


class StrEnum(str, Enum):
    """Enum with possibility to be a query param serializable"""

    # also for params serialization, make `print` to show only value which complicates debugging a little
    def __str__(self):
        return self.value


class App(IntEnum):
    """App enum. Add new member, when missing"""

    # predefined
    CS2 = 730
    CSGO = CS2  # alias

    DOTA2 = 570
    H1Z1 = 433850
    RUST = 252490
    TF2 = 440
    PUBG = 578080

    STEAM = 753

    @classmethod
    def extend(cls, name: str, value: int) -> "App":
        return extend_enum(cls, name, value)

    @classmethod
    def _generate_name(cls, value) -> str:
        return f"{cls.__name__}_{value}"

    @classmethod
    def _missing_(cls, value: int):
        return cls.extend(cls._generate_name(value), value)  # add new member when missing

    @property
    def app_id(self) -> int:
        return self.value


class AppContext(Enum):
    """
    Combination of `App` and context (sub-inventory of the app)

    .. seealso:: https://dev.doctormckay.com/topic/332-identifying-steam-items/
    """

    # predefined
    CS2 = App.CS2, 2
    CSGO = CS2  # alias

    DOTA2 = App.DOTA2, 2
    H1Z1 = App.H1Z1, 2
    RUST = App.RUST, 2
    TF2 = App.TF2, 2
    PUBG = App.PUBG, 2

    STEAM_GIFTS = App.STEAM, 1
    STEAM_COMMUNITY = App.STEAM, 6
    STEAM_REWARDS = App.STEAM, 7  # item rewards

    @classmethod
    def extend(cls, name: str, value: tuple[App | int, int]) -> "AppContext":
        # for case when passed app id instead of an App enum, e.g. AppContext((730, 2))
        with_enum = (App(value[0]), value[1])
        return extend_enum(cls, name, with_enum)

    @classmethod
    def _generate_name(cls, value: tuple[App, int]) -> str:
        return f"{cls.__name__}_{value[0]}_{value[1]}"

    @classmethod
    def _missing_(cls, value: tuple[App | int, int]):
        with_enum = (App(value[0]), value[1])
        return cls.extend(cls._generate_name(with_enum), with_enum)

    @property
    def app(self) -> App:
        return self.value[0]

    @property
    def app_id(self) -> int:
        return self.value[0].value

    @property
    def context(self) -> int:
        return self.value[1]


class Currency(IntEnum):  # already params serializable
    """
    Steam currency enum.

    .. seealso:: https://partner.steamgames.com/doc/store/pricing/currencies
    """

    USD = 1  # UnitedStates Dollar
    GBP = 2  # United Kingdom Pound
    # EURO = 3  # European Union Euro
    EUR = 3  # European Union Euro
    CHF = 4  # Swiss Francs
    RUB = 5  # Russian Rouble
    PLN = 6  # Polish Złoty
    BRL = 7  # Brazilian Reals
    JPY = 8  # Japanese Yen
    NOK = 9  # Norwegian Krone
    IDR = 10  # Indonesian Rupiah
    MYR = 11  # Malaysian Ringgit
    PHP = 12  # Philippine Peso
    SGD = 13  # Singapore Dollar
    THB = 14  # Thai Baht
    VND = 15  # Vietnamese Dong
    KRW = 16  # South KoreanWon
    TRY = 17  # Turkish Lira
    UAH = 18  # Ukrainian Hryvnia
    MXN = 19  # Mexican Peso
    CAD = 20  # Canadian Dollars
    AUD = 21  # Australian Dollars
    NZD = 22  # New Zealand Dollar
    CNY = 23  # Chinese Renminbi (yuan)
    INR = 24  # Indian Rupee
    CLP = 25  # Chilean Peso
    PEN = 26  # Peruvian Sol
    COP = 27  # Colombian Peso
    ZAR = 28  # South AfricanRand
    HKD = 29  # Hong KongDollar
    TWD = 30  # New TaiwanDollar
    SAR = 31  # Saudi Riyal
    AED = 32  # United ArabEmirates Dirham
    # SEK = 33  # Swedish Krona
    ARS = 34  # Argentine Peso
    ILS = 35  # Israeli NewShekel
    # BYN = 36  # Belarusian Ruble
    KZT = 37  # Kazakhstani Tenge
    KWD = 38  # Kuwaiti Dinar
    QAR = 39  # Qatari Riyal
    CRC = 40  # Costa Rican Colón
    UYU = 41  # Uruguayan Peso
    # BGN = 42  # Bulgarian Lev
    # HRK = 43  # Croatian Kuna
    # CZK = 44  # Czech Koruna
    # DKK = 45  # Danish Krone
    # HUF = 46  # Hungarian Forint
    # RON = 47  # Romanian Leu


class Language(StrEnum):
    """
    Steam languages.

    .. seealso:: https://partner.steamgames.com/doc/store/localization/languages
    """

    ARABIC = "arabic"
    BULGARIAN = "bulgarian"
    SIMPLIFIED_CHINESE = "schinese"
    TRADITIONAL_CHINESE = "tchinese"
    CZECH = "czech"
    DANISH = "danish"
    DUTCH = "dutch"
    ENGLISH = "english"
    FINNISH = "finnish"
    FRENCH = "french"
    GERMAN = "german"
    GREEK = "greek"
    HUNGARIAN = "hungarian"
    ITALIAN = "italian"
    JAPANESE = "japanese"
    KOREAN = "koreana"
    NORWEGIAN = "norwegian"
    POLISH = "polish"
    PORTUGUESE = "portuguese"
    PORTUGUESE_BRAZIL = "brazilian"
    ROMANIAN = "romanian"
    RUSSIAN = "russian"
    SPANISH = "spanish"
    SPANISH_LATIN_AMERICAN = "latam"
    SWEDISH = "swedish"
    THAI = "thai"
    TURKISH = "turkish"
    UKRAINIAN = "ukrainian"
    VIETNAMESE = "vietnamese"


class TradeOfferStatus(Enum):
    INVALID = 1
    ACTIVE = 2
    ACCEPTED = 3
    COUNTERED = 4
    EXPIRED = 5
    CANCELED = 6
    DECLINED = 7
    INVALID_ITEMS = 8
    CONFIRMATION_NEED = 9
    CANCELED_BY_SECONDARY_FACTOR = 10
    STATE_IN_ESCROW = 11


# https://github.com/DoctorMcKay/node-steamcommunity/blob/master/resources/EConfirmationType.js
class ConfirmationType(Enum):
    UNKNOWN = 1
    TRADE = 2
    LISTING = 3
    API_KEY = 4  # TODO find api key value

    @classmethod
    def get(cls, v: int) -> "ConfirmationType":
        try:
            return cls(v)
        except ValueError:
            return cls.UNKNOWN


class MarketListingStatus(Enum):
    NEED_CONFIRMATION = 17
    ACTIVE = 1


class MarketHistoryEventType(Enum):
    LISTING_CREATED = 1
    LISTING_CANCELED = 2
    LISTING_SOLD = 3
    LISTING_PURCHASED = 4


# TODO Maybe I can create class with __getattribute__ and then build url trough magic method calls
_API_BASE = URL("https://api.steampowered.com")  # nah
_v = "v1"


class STEAM_URL:
    COMMUNITY = URL("https://steamcommunity.com")  # use this domain in methods
    STORE = URL("https://store.steampowered.com")
    LOGIN = URL("https://login.steampowered.com")
    HELP = URL("https://help.steampowered.com")
    STATIC = URL("https://community.akamai.steamstatic.com")
    # specific
    MARKET = COMMUNITY / "market/"
    TRADE = COMMUNITY / "tradeoffer"

    class API:
        BASE = _API_BASE

        # interfaces
        class IEconService:
            _Base = _API_BASE / "IEconService"

            GetTradeHistory = _Base / "GetTradeHistory" / _v
            GetTradeHoldDurations = _Base / "GetTradeHoldDurations" / _v
            GetTradeOffer = _Base / "GetTradeOffer" / _v
            GetTradeOffers = _Base / "GetTradeOffers" / _v
            GetTradeOffersSummary = _Base / "GetTradeOffersSummary" / _v
            GetTradeStatus = _Base / "GetTradeStatus" / _v

        class IAuthService:
            _Base = _API_BASE / "IAuthenticationService"

            BeginAuthSessionViaCredentials = _Base / "BeginAuthSessionViaCredentials" / _v
            GetPasswordRSAPublicKey = _Base / "GetPasswordRSAPublicKey" / _v
            UpdateAuthSessionWithSteamGuardCode = _Base / "UpdateAuthSessionWithSteamGuardCode" / _v
            PollAuthSessionStatus = _Base / "PollAuthSessionStatus" / _v
            GenerateAccessTokenForApp = _Base / "GenerateAccessTokenForApp" / _v


T_PARAMS: TypeAlias = Mapping[str, int | str | float]
T_PAYLOAD: TypeAlias = Mapping[str, str | int | float | bool | None | list | Mapping]
T_HEADERS: TypeAlias = Mapping[str, str]


class EnumWithMultipleValues(Enum):
    """Author: ChatGPT"""

    def __new__(cls, *values):
        # The first value is considered the primary value for the enum member
        obj = object.__new__(cls)
        obj._value_ = values[0]
        # Store all the additional values in a class-level dictionary
        if not hasattr(cls, "_alt_map"):
            cls._alt_map = {}
        for value in values:
            cls._alt_map[value] = obj
        return obj

    @classmethod
    def _missing_(cls, value):
        # Handle cases where the value doesn't directly map to a member
        return cls._alt_map.get(value, super()._missing_(value))


# https://github.com/DoctorMcKay/node-steamcommunity/blob/master/resources/EResult.js
class EResult(EnumWithMultipleValues):
    """
    `success` field in response data from Steam.

    .. seealso:: https://steamerrors.com
    """

    UNKNOWN = None  # special case

    INVALID = 0
    OK = 1, True  # due to Steam
    FAIL = 2
    NO_CONNECTION = 3
    INVALID_PASSWORD = 5
    LOGGED_IN_ELSEWHERE = 6
    INVALID_PROTOCOL_VER = 7
    INVALID_PARAM = 8
    FILE_NOT_FOUND = 9
    BUSY = 10
    INVALID_STATE = 11
    INVALID_NAME = 12
    INVALID_EMAIL = 13
    DUPLICATE_NAME = 14
    ACCESS_DENIED = 15
    TIMEOUT = 16
    BANNED = 17
    ACCOUNT_NOT_FOUND = 18
    INVALID_STEAM_ID = 19
    SERVICE_UNAVAILABLE = 20
    NOT_LOGGED_ON = 21
    PENDING = 22
    ENCRYPTION_FAILURE = 23
    INSUFFICIENT_PRIVILEGE = 24
    LIMIT_EXCEEDED = 25
    REVOKED = 26
    EXPIRED = 27
    ALREADY_REDEEMED = 28
    DUPLICATE_REQUEST = 29
    ALREADY_OWNED = 30
    IP_NOT_FOUND = 31
    PERSIST_FAILED = 32
    LOCKING_FAILED = 33
    LOGON_SESSION_REPLACED = 34
    CONNECT_FAILED = 35
    HANDSHAKE_FAILED = 36
    IO_FAILURE = 37
    REMOTE_DISCONNECT = 38
    SHOPPING_CART_NOT_FOUND = 39
    BLOCKED = 40
    IGNORED = 41
    NO_MATCH = 42
    ACCOUNT_DISABLED = 43
    SERVICE_READ_ONLY = 44
    ACCOUNT_NOT_FEATURED = 45
    ADMINISTRATOR_OK = 46
    CONTENT_VERSION = 47
    TRY_ANOTHER_CM = 48
    PASSWORD_REQUIRED_TO_KICK_SESSION = 49
    ALREADY_LOGGED_IN_ELSEWHERE = 50
    SUSPENDED = 51
    CANCELLED = 52
    DATA_CORRUPTION = 53
    DISK_FULL = 54
    REMOTE_CALL_FAILED = 55
    # PASSWORD_NOT_SET= 56 // removed "renamed to PasswordUnset"
    PASSWORD_UNSET = 56
    EXTERNAL_ACCOUNT_UNLINKED = 57
    PSN_TICKET_INVALID = 58
    EXTERNAL_ACCOUNT_ALREADY_LINKED = 59
    REMOTE_FILE_CONFLICT = 60
    ILLEGAL_PASSWORD = 61
    SAME_AS_PREVIOUS_VALUE = 62
    ACCOUNT_LOGON_DENIED = 63
    CANNOT_USE_OLD_PASSWORD = 64
    INVALID_LOGIN_AUTH_CODE = 65
    # ACCOUNT_LOGON_DENIED_NO_MAIL_SENT= 66 // removed "renamed to AccountLogonDeniedNoMail"
    ACCOUNT_LOGON_DENIED_NO_MAIL = 66
    HARDWARE_NOT_CAPABLE_OF_IPT = 67
    IPT_INIT_ERROR = 68
    PARENTAL_CONTROL_RESTRICTED = 69
    FACEBOOK_QUERY_ERROR = 70
    EXPIRED_LOGIN_AUTH_CODE = 71
    IP_LOGIN_RESTRICTION_FAILED = 72
    # ACCOUNT_LOCKED= 73 // removed "renamed to AccountLockedDown"
    ACCOUNT_LOCKED_DOWN = 73
    ACCOUNT_LOGON_DENIED_VERIFIED_EMAIL_REQUIRED = 74
    NO_MATCHING_URL = 75
    BAD_RESPONSE = 76
    REQUIRE_PASSWORD_RE_ENTRY = 77
    VALUE_OUT_OF_RANGE = 78
    UNEXPECTED_ERROR = 79
    DISABLED = 80
    INVALID_CEG_SUBMISSION = 81
    RESTRICTED_DEVICE = 82
    REGION_LOCKED = 83
    RATE_LIMIT_EXCEEDED = 84
    # ACCOUNT_LOGON_DENIED_NEED_TWO_FACTOR_CODE= 85 // removed "renamed to AccountLoginDeniedNeedTwoFactor"
    ACCOUNT_LOGIN_DENIED_NEED_TWO_FACTOR = 85
    # ITEM_OR_ENTRY_HAS_BEEN_DELETED= 86 // removed "renamed to ItemDeleted"
    ITEM_DELETED = 86
    ACCOUNT_LOGIN_DENIED_THROTTLE = 87
    TWO_FACTOR_CODE_MISMATCH = 88
    TWO_FACTOR_ACTIVATION_CODE_MISMATCH = 89
    # ACCOUNT_ASSOCIATED_TO_MULTIPLE_PLAYERS= 90 // removed "renamed to AccountAssociatedToMultiplePartners"
    ACCOUNT_ASSOCIATED_TO_MULTIPLE_PARTNERS = 90
    NOT_MODIFIED = 91
    # NO_MOBILE_DEVICE_AVAILABLE= 92 // removed "renamed to NoMobileDevice"
    NO_MOBILE_DEVICE = 92
    # TIME_IS_OUT_OF_SYNC= 93 // removed "renamed to TimeNotSynced"
    TIME_NOT_SYNCED = 93
    SMS_CODE_FAILED = 94
    # TOO_MANY_ACCOUNTS_ACCESS_THIS_RESOURCE= 95 // removed "renamed to AccountLimitExceeded"
    ACCOUNT_LIMIT_EXCEEDED = 95
    ACCOUNT_ACTIVITY_LIMIT_EXCEEDED = 96
    PHONE_ACTIVITY_LIMIT_EXCEEDED = 97
    REFUND_TO_WALLET = 98
    EMAIL_SEND_FAILURE = 99
    NOT_SETTLED = 100
    NEED_CAPTCHA = 101
    GSLT_DENIED = 102
    GS_OWNER_DENIED = 103
    INVALID_ITEM_TYPE = 104
    IP_BANNED = 105
    GSLT_EXPIRED = 106
    INSUFFICIENT_FUNDS = 107
    TOO_MANY_PENDING = 108
    NO_SITE_LICENSES_FOUND = 109
    WG_NETWORK_SEND_EXCEEDED = 110
    ACCOUNT_NOT_FRIENDS = 111
    LIMITED_USER_ACCOUNT = 112
    CANT_REMOVE_ITEM = 113
    ACCOUNT_HAS_BEEN_DELETED = 114
    ACCOUNT_HAS_AN_EXISTING_USER_CANCELLED_LICENSE = 115
    DENIED_DUE_TO_COMMUNITY_COOLDOWN = 116
    NO_LAUNCHER_SPECIFIED = 117
    MUST_AGREE_TO_SSA = 118
    CLIENT_NO_LONGER_SUPPORTED = 119


TRADABLE_AFTER_DATE_FORMAT = "Tradable After %b %d, %Y (%H:%M:%S) %Z"
